#define _XOPEN_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_math.h>
#include <float.h>
#include <sys/time.h>
#include <unistd.h>
#include <CL/cl.h>
#include <string.h>
#include <sys/stat.h>
#include <assert.h>


struct System{
    int L;
    int N;  // 2*L
    double/**/ **thresholds;
    double/**/ **nodes;
    double/**/ *pathscape;
    double/**/ **pathscape_bonds;
};

struct Params{
    int size;
    int i_row;
    int i_col;
    int col;
    int row;
};

double walltime (void){
	static struct timeval t;
	gettimeofday ( &t, NULL );
	return ( t.tv_sec + 1e-6 * t.tv_usec );
}


void initSys(struct System *sys, int l, int seed){

    int L = l;  // L = height of system
    int N = l;  // N = width of system
    sys->L = L;
    sys->N = N;
    sys->pathscape = (double/**/*)calloc(N*(L+1), sizeof(sys->pathscape));
    sys->pathscape_bonds = (double/**/**)calloc(N*(L+1), sizeof(sys->pathscape_bonds));
    sys->thresholds = (double/**/**)calloc(N*L, sizeof(sys->thresholds));
    sys->nodes = (double/**/**)calloc(2*N, sizeof(sys->nodes));
    for(int i=0; i<N*L; i++)
	sys->thresholds[i] = (double/**/*)calloc(2, sizeof(sys->thresholds[i]));
    for(int i=0; i<N*(L+1); i++){
	sys->pathscape_bonds[i] = (double/**/*)calloc(2, sizeof(sys->pathscape_bonds[i]));   
	sys->pathscape[i] = DBL_MAX;
    }
    for(int i=0; i<2*N; i++)
	sys->nodes[i] = (double/**/*)calloc(N*L, sizeof(sys->nodes[i]));


    gsl_rng *r = gsl_rng_alloc(gsl_rng_mt19937);
    gsl_rng_set(r, seed);
    for(int i=0; i<N*L; i++){
	sys->thresholds[i][0] = gsl_rng_uniform(r);
	sys->thresholds[i][1] = gsl_rng_uniform(r);
    }
    gsl_rng_free(r);
}

void freeSys(struct System *sys){

    for(int i=0; i<sys->N*sys->L; i++)
	free(sys->thresholds[i]);

    for(int i=0; i<sys->N*(sys->L+1); i++)
	free(sys->pathscape_bonds[i]);    
    
    for(int i=0; i<2*sys->N; i++)
	free(sys->nodes[i]);

    free(sys->thresholds);
    free(sys->nodes);
    free(sys->pathscape);
    free(sys->pathscape_bonds);

}

inline double/**/ minNeighbour3(double/**/ s1, double/**/ s2, double/**/ s3, int *k){

    double/**/ m;
    
    *k = (s1 < s2) ? 1 : 2;
    m = (s1 < s2) ? s1 : s2;
    *k = (s3 < m) ? 3 : *k;
    m = (s3 < m) ? s3 : m;

    return m;
}
inline double/**/ minNeighbour4(double/**/ s1, double/**/ s2, double/**/ s3, double/**/ s4, int *k){

    double/**/ m;

    *k = (s1 < s2) ? 1 : 2;
    m = (s1 < s2) ? s1 : s2;
    *k = (s3 < m) ? 3 : *k;
    m = (s3 < m) ? s3 : m;
    *k = (s4 < m) ? 4 : *k;
    m = (s4 < m) ? s4 : m;

    return m;
}
void calculateOptimalPathscapeNEWTHATDIDNTREALLYHELP(double/**/ **nodes, 
			       double/**/ **temp_thresholds, 
			       int **bonds, int L, double/**/ tol, int row){


    int N = 2*L;
    int l = L/2;
    int counter = 0;
    enum{ LEFT = 1, BELOW, ABOVE, RIGHT};

    for(int i=0; i<2; i++){

	double/**/ sum, old_sum, s1, s2, s3, s4;
	int i_node, n, t;
	
	sum = 0;
	old_sum = 1;
	
	// i_node should not be updated (locked to zero)
	i_node = (i == 0) ? row*N+l : row*N+L+l;
	nodes[i][i_node] = 0;
	//while(fabs(sum-old_sum) > tol){
	counter = 0;
	while(counter < 10){
	    old_sum = sum;
	    sum = 0;
	    

	    // j == 0
	    for(int k=0; k<N-1; k++){
		s3 = nodes[i][ k+N   ] + temp_thresholds[ k   ][1];
		s4 = nodes[i][ k+N+1 ] + temp_thresholds[ k+1 ][0];
		nodes[i][ k ] = (s3 < s4) ? s3 : s4;
		nodes[i][ k ] *= (k != i_node);
		bonds[i][ k ] = (nodes[i][ k ] > 0) * (s3 < s4 ? ABOVE : RIGHT);
		sum += nodes[i][ k ];
	    }

	    // j == 0, k == N-1
	    nodes[i][ N-1 ] = (nodes[i][ 2*N-1 ] + temp_thresholds[ N-1 ][1]) * (N-1 != i_node); 
	    bonds[i][ N-1 ] = (nodes[i][ N-1 ] > 0) * ABOVE;
	    sum += nodes[i][ N-1 ];

	    for(int j=1; j<L; j++){

		// k == 0
		n = j*N;
		s2 = nodes[i][ j*N-N   ] + temp_thresholds[ j*N-N ][1];
		s3 = nodes[i][ j*N+N   ] + temp_thresholds[ j*N   ][1];
		s4 = nodes[i][ j*N+N+1 ] + temp_thresholds[ j*N+1 ][0];
		nodes[i][ n ] = minNeighbour3(s2, s3, s4, &t) * (n != i_node);
		bonds[i][ n ] = (nodes[i][ n ] > 0) * (t+1);
		sum += nodes[i][ n ];
		
		for(int k=1; k<N-1; k++){

		    n = j*N+k;
		    s1 = nodes[i][ j*N+k-N-1 ] + temp_thresholds[ j*N+k-N ][0];
		    s2 = nodes[i][ j*N+k-N   ] + temp_thresholds[ j*N+k-N ][1];
		    s3 = nodes[i][ j*N+k+N   ] + temp_thresholds[ j*N+k   ][1];
		    s4 = nodes[i][ j*N+k+N+1 ] + temp_thresholds[ j*N+k+1 ][0];
		    
		    nodes[i][ n ] = minNeighbour4(s1, s2, s3, s4, &t) * (n != i_node);
		    bonds[i][ n ] = (nodes[i][ n ] > 0) * t;
		    sum += nodes[i][ n ];
		}

		// k == N-1
		n = j*N+N-1;
		s1 = nodes[i][ j*N-2     ] + temp_thresholds[ j*N-1   ][0];
		s2 = nodes[i][ j*N+-1    ] + temp_thresholds[ j*N-1   ][1];
		s3 = nodes[i][ j*N+2*N-1 ] + temp_thresholds[ j*N+N-1 ][1];
		nodes[i][ n ] = minNeighbour3(s1, s2, s3, &t) * (n != i_node);
		bonds[i][ n ] = (nodes[i][ n ] > 0) * t;
		sum += nodes[i][ n ];

	    }

	    // j == L
	    // k == 0
	    nodes[i][ L*N ] = (nodes[i][ L*N-N   ] + temp_thresholds[ L*N-N ][1]) * (L*N != i_node);
	    bonds[i][ L*N ] = (nodes[i][ L*N ] > 0) * BELOW;
	    sum += nodes[i][ L*N ];

	    for(int k=1; k<N; k++){
		n = L*N+k;
		s1 = nodes[i][ L*N+k-N-1 ] + temp_thresholds[ L*N+k-N ][0];
		s2 = nodes[i][ L*N+k-N   ] + temp_thresholds[ L*N+k-N ][1];
		nodes[i][ n ] = (s1 < s2) ? s1 : s2;
		nodes[i][ n ] *= (n != i_node);
		bonds[i][ n ] = (nodes[i][ n ] > 0) * (s1 < s2 ? LEFT : BELOW);
		sum += nodes[i][ n ];
	    }

	    counter += !(fabs(sum-old_sum) > tol);
	    counter *= !(fabs(sum-old_sum) > tol);
	}

	// As source nodes close to each other have similar distances to
	// the other nodes, the next run through starts from the results
	// of this one. No need for priming pathscape, as this is much
	// faster (factor of ~15 for 16x16). 
	// Need to adjust this for 4 threads
	// ASSUMPTION: The first call to calculateOptimalPathscape
	// is for row 0, followed by rows 1, 2, ..., L+1.
	if(i == 0 && row == 0)
	    for(int j=0; j<N*(L+1); j++){
		nodes[1][j] = nodes[0][j];
		bonds[1][j] = bonds[0][j];
	    }
    }
}




void findFracturescape(struct System *sys){
    
    int L = sys->L, N = 2*L, l = L/2;
    int k1, k2,
	*conv_col,
	**bonds, **temp_bonds0, **temp_bonds1;
    double/**/ *sp_nodes, 
	   **nodes, **temp_nodes, **temp_thresholds;

    conv_col = (int*)calloc(N, sizeof(conv_col));
    for(int i=0; i<l; i++)
	conv_col[i] = i+l;
    for(int i=l; i<l+L; i++)
	conv_col[i] = i-l;
    for(int i=l+L; i<N; i++)
	conv_col[i] = i-l-L;

    nodes = (double/**/**)calloc(2, sizeof(nodes));
    temp_nodes = (double/**/**)calloc(L+1, sizeof(temp_nodes));
    sp_nodes = (double/**/*)calloc(N*(L+1), sizeof(sp_nodes));
    bonds = (int**)calloc(2, sizeof(bonds));
    temp_bonds0 = (int**)calloc(L+1, sizeof(temp_bonds0));
    temp_bonds1 = (int**)calloc(L+1, sizeof(temp_bonds1));
    temp_thresholds = (double/**/**)calloc(N*L, sizeof(temp_thresholds));
    for(int i=0; i<2; i++){
    	nodes[i] = (double/**/*)calloc(N*(L+1), sizeof(nodes[i]));
    	bonds[i] = (int*)calloc(N*(L+1), sizeof(bonds[i]));
    }
    for(int i=0; i<L+1; i++){
    	temp_nodes[i] = (double/**/*)calloc(L*(L+1), sizeof(temp_nodes[i]));
    	temp_bonds0[i] = (int*)calloc(L*(L+1), sizeof(temp_bonds0[i]));
    	temp_bonds1[i] = (int*)calloc(L*(L+1), sizeof(temp_bonds1[i]));
    }


    for(int i=0; i<L; i++){
    	for(int j=0; j<N; j++){
    	    temp_thresholds[ i*N+j ] = (double/**/*)calloc(2, sizeof(temp_thresholds[i*N+j]));
    	    temp_thresholds[ i*N+j ][0] = sys->thresholds[ i*L+conv_col[j] ][0];
    	    temp_thresholds[ i*N+j ][1] = sys->thresholds[ i*L+conv_col[j] ][1];
    	}
    }


    // Extract optimal pathscape to sys

    //calculateOptimalPathscapeNEWTHATDIDNTREALLYHELP(nodes, temp_thresholds, bonds, L, 1e-10, 0);

    for(int i=0; i<L+1; i++){
    	// i is the current row that the spanning path is calculated for
    	calculateOptimalPathscapeNEWTHATDIDNTREALLYHELP(nodes, temp_thresholds, bonds, L, 1e-10, i);


    	// sp_nodes[i][:] are set to the sum of nodes[0][:] and new_nodes[1][:],
    	// i.e. total threshold for spanning paths going through the different nodes
    	for(int j=0; j<N*(L+1); j++){
    	    sp_nodes[j] = (nodes[0][j] + nodes[1][j]);
    	}

    	// temp_nodes[] is of same size as pathscape.
    	// Set node values to the minimum of paths going through columns k and k'
    	for(int j=0; j<L+1; j++)
    	    for(int k=0; k<L; k++){
    		k1 = k+l;
    		k2 = (k < l) ? k+L+l : k-l;

    		temp_bonds0[i][j*L+k] = sp_nodes[j*N+k1] < sp_nodes[j*N+k2] ?
    		    bonds[0][j*N+k1] : bonds[0][j*N+k2];

    		temp_bonds1[i][j*L+k] = sp_nodes[j*N+k1] < sp_nodes[j*N+k2] ?
    		    bonds[1][j*N+k1] : bonds[1][j*N+k2];

    		temp_nodes[i][j*L+k] = sp_nodes[j*N+k1] < sp_nodes[j*N+k2] ?
    		    sp_nodes[j*N+k1] : sp_nodes[j*N+k2];
    	    }
    }

    // temp_nodes and temp_bonds take a lot of memory.
    // Write directly to pathscape after finding sp_nodes
    // instead of going via temp_nodes and temp_bonds?
    for(int i=0; i<L+1; i++){
    	for(int j=0; j<L*(L+1); j++){
    	    if(fabs(temp_nodes[i][j]) > 1e-6){
    		sys->pathscape_bonds[j][0] =
    		    sys->pathscape[j] < temp_nodes[i][j] ?
    		    sys->pathscape_bonds[j][0] : temp_bonds0[i][j];

    		sys->pathscape_bonds[j][1] =
    		    sys->pathscape[j] < temp_nodes[i][j] ?
    		    sys->pathscape_bonds[j][1] : temp_bonds1[i][j];

    		sys->pathscape[j] =
    		    sys->pathscape[j] < temp_nodes[i][j] ?
    		    sys->pathscape[j] : temp_nodes[i][j];
    	    }
	    //else
	    //	printf("Low val for i = %d, j = %d\n", i, j);
    	}
    }

    free(conv_col);
    for(int i=0; i<N*L; i++)
	free(temp_thresholds[i]);
    for(int i=0; i<2; i++){
	free(bonds[i]);
	free(nodes[i]);
    }
    for(int i=0; i<L+1; i++){
	free(temp_bonds0[i]);
	free(temp_bonds1[i]);
	free(temp_nodes[i]);
    }
    free(sp_nodes);
    free(temp_thresholds);
    free(bonds);
    free(temp_bonds0);
    free(temp_bonds1);
    free(nodes);
    free(temp_nodes);
}
void prepareNodes(struct System *sys, double/**/ *nodes,
		  double/**/ *vthresholds, double/**/ *dthresholds){

    int L=sys->L, N=2*L, l=L/2;
    int *conv_col;
    int rows=L+3;
    int cols=N+2;
    
    // nodes is of size (L+3)*(N+2) = rows*cols
    // v/dthresholds is of size (L+2)*(N+2)

    for(int i=0; i<cols; i++){
	nodes[i] = DBL_MAX;
	nodes[(rows-1)*cols+i] = DBL_MAX;
	vthresholds[i] = DBL_MAX;
	vthresholds[(L+1)*cols+i] = DBL_MAX;
	dthresholds[i] = DBL_MAX;
	dthresholds[(L+1)*cols+i] = DBL_MAX;
    }
    for(int i=0; i<rows; i++){
    	nodes[i*cols] = DBL_MAX;
    	nodes[(i+1)*cols-1] = DBL_MAX;
    }
    for(int i=0; i<rows-1; i++){
	vthresholds[i*cols] = DBL_MAX;
	vthresholds[(i+1)*cols-1] = DBL_MAX;
	dthresholds[i*cols] = DBL_MAX;
	dthresholds[(i+1)*cols-1] = DBL_MAX;
    }
    
    conv_col = (int*)calloc(N, sizeof(conv_col));
    for(int i=0; i<l; i++)
	conv_col[i] = i+l;
    for(int i=l; i<l+L; i++)
	conv_col[i] = i-l;
    for(int i=l+L; i<N; i++)
	conv_col[i] = i-l-L;


    for(int i=0; i<L; i++){
    	for(int j=0; j<N; j++){
	    vthresholds[ (i+1)*cols+j+1 ] = sys->thresholds[ i*L+conv_col[j] ][0];
    	    dthresholds[ (i+1)*cols+j+1 ] = sys->thresholds[ i*L+conv_col[j] ][1];
    	}
    }

    free(conv_col);

}

void writeToGPUFracturescape(int L, double/**/ *fracturescape, int **bonds,
			     double/**/ *nodes1, double/**/ *nodes2,
			     int *bonds1, int *bonds2,
			     double/**/ *temp_fracturescape, int **temp_bonds){
    
 
    int l = L/2;
    int N = 2*L+2;
    int k1, k2;
    double comb1, comb2;
    // Using combined as a list somehow crashes when freeing with double
    // Don't need to us it as a list anyway, just use two temporary variables,
    // comb1 and comb2
    //double/**/ *combined = (double/**/*)calloc(2*L*(L+1), sizeof(combined));
    for(int j=0; j<L+1; j++)
	for(int k=0; k<L; k++){
	    k1 = k+l;
	    k2 = (k < l) ? k+L+l : k-l;
	    k1++;  //Extra column due to 'wall', extra row is done directly on j
	    k2++;

	    comb1 = nodes1[(j+1)*N+k1]+nodes2[(j+1)*N+k1];
	    comb2 = nodes1[(j+1)*N+k2]+nodes2[(j+1)*N+k2];

	    temp_bonds[0][j*L+k] = comb1 < comb2 ?
		bonds1[(j+1)*N+k1] : bonds1[(j+1)*N+k2];

	    temp_bonds[1][j*L+k] = comb1 < comb2 ?
		bonds2[(j+1)*N+k1] : bonds2[(j+1)*N+k2];

	    temp_fracturescape[j*L+k] = comb1 < comb2 ?
		comb1 : comb2;

	}

    for(int j=0; j<L+1; j++)
	for(int k=0; k<L; k++){
	    bonds[0][j*L+k] = temp_fracturescape[j*L+k] < fracturescape[j*L+k] ? 
		  temp_bonds[0][j*L+k] : bonds[0][j*L+k];
	    bonds[1][j*L+k] = temp_fracturescape[j*L+k] < fracturescape[j*L+k] ? 
		  temp_bonds[1][j*L+k] : bonds[1][j*L+k];
	    fracturescape[j*L+k] = temp_fracturescape[j*L+k] < fracturescape[j*L+k] ? 
                  temp_fracturescape[j*L+k] : fracturescape[j*L+k];
	    
	}				  
}



char * load_program_source(const char *filename){ 
	
	struct stat statbuf;
	FILE *fh; 
	char *source; 
	
	fh = fopen(filename, "r");
	if (fh == 0)
		return 0; 
	
	stat(filename, &statbuf);
	source = (char *) malloc(statbuf.st_size + 1);
	fread(source, statbuf.st_size, 1, fh);
	source[statbuf.st_size] = '\0'; 
	
	return source; 
}

char *print_cl_errstring(cl_int err){
    switch (err) {
        case CL_SUCCESS:                          return strdup("Success!");
        case CL_DEVICE_NOT_FOUND:                 return strdup("Device not found.");
        case CL_DEVICE_NOT_AVAILABLE:             return strdup("Device not available");
        case CL_COMPILER_NOT_AVAILABLE:           return strdup("Compiler not available");
        case CL_MEM_OBJECT_ALLOCATION_FAILURE:    return strdup("Memory object allocation failure");
        case CL_OUT_OF_RESOURCES:                 return strdup("Out of resources");
        case CL_OUT_OF_HOST_MEMORY:               return strdup("Out of host memory");
        case CL_PROFILING_INFO_NOT_AVAILABLE:     return strdup("Profiling information not available");
        case CL_MEM_COPY_OVERLAP:                 return strdup("Memory copy overlap");
        case CL_IMAGE_FORMAT_MISMATCH:            return strdup("Image format mismatch");
        case CL_IMAGE_FORMAT_NOT_SUPPORTED:       return strdup("Image format not supported");
        case CL_BUILD_PROGRAM_FAILURE:            return strdup("Program build failure");
        case CL_MAP_FAILURE:                      return strdup("Map failure");
        case CL_INVALID_VALUE:                    return strdup("Invalid value");
        case CL_INVALID_DEVICE_TYPE:              return strdup("Invalid device type");
        case CL_INVALID_PLATFORM:                 return strdup("Invalid platform");
        case CL_INVALID_DEVICE:                   return strdup("Invalid device");
        case CL_INVALID_CONTEXT:                  return strdup("Invalid context");
        case CL_INVALID_QUEUE_PROPERTIES:         return strdup("Invalid queue properties");
        case CL_INVALID_COMMAND_QUEUE:            return strdup("Invalid command queue");
        case CL_INVALID_HOST_PTR:                 return strdup("Invalid host pointer");
        case CL_INVALID_MEM_OBJECT:               return strdup("Invalid memory object");
        case CL_INVALID_IMAGE_FORMAT_DESCRIPTOR:  return strdup("Invalid image format descriptor");
        case CL_INVALID_IMAGE_SIZE:               return strdup("Invalid image size");
        case CL_INVALID_SAMPLER:                  return strdup("Invalid sampler");
        case CL_INVALID_BINARY:                   return strdup("Invalid binary");
        case CL_INVALID_BUILD_OPTIONS:            return strdup("Invalid build options");
        case CL_INVALID_PROGRAM:                  return strdup("Invalid program");
        case CL_INVALID_PROGRAM_EXECUTABLE:       return strdup("Invalid program executable");
        case CL_INVALID_KERNEL_NAME:              return strdup("Invalid kernel name");
        case CL_INVALID_KERNEL_DEFINITION:        return strdup("Invalid kernel definition");
        case CL_INVALID_KERNEL:                   return strdup("Invalid kernel");
        case CL_INVALID_ARG_INDEX:                return strdup("Invalid argument index");
        case CL_INVALID_ARG_VALUE:                return strdup("Invalid argument value");
        case CL_INVALID_ARG_SIZE:                 return strdup("Invalid argument size");
        case CL_INVALID_KERNEL_ARGS:              return strdup("Invalid kernel arguments");
        case CL_INVALID_WORK_DIMENSION:           return strdup("Invalid work dimension");
        case CL_INVALID_WORK_GROUP_SIZE:          return strdup("Invalid work group size");
        case CL_INVALID_WORK_ITEM_SIZE:           return strdup("Invalid work item size");
        case CL_INVALID_GLOBAL_OFFSET:            return strdup("Invalid global offset");
        case CL_INVALID_EVENT_WAIT_LIST:          return strdup("Invalid event wait list");
        case CL_INVALID_EVENT:                    return strdup("Invalid event");
        case CL_INVALID_OPERATION:                return strdup("Invalid operation");
        case CL_INVALID_GL_OBJECT:                return strdup("Invalid OpenGL object");
        case CL_INVALID_BUFFER_SIZE:              return strdup("Invalid buffer size");
        case CL_INVALID_MIP_LEVEL:                return strdup("Invalid mip-map level");
        default:                                  return strdup("Unknown");
    }
}



#define DEVICEID 0



int clError(cl_int cl_err, const char *str){

    if(cl_err != CL_SUCCESS){
	printf("%s failed: %s\n", str, print_cl_errstring(cl_err));
	return 1;
    }
 
    return 0;
}

int main(void){


    for(int run_throughs = 8; run_throughs < 9; run_throughs++){
    cl_platform_id *cl_platform;   // Platform ID
    cl_device_id   *cl_device;     // Device ID
    cl_int          cl_err;        // Error code
    cl_uint         num_platforms;
    cl_uint         num_devices;
    cl_event        event;
    cl_ulong        start_time, end_time;
    

    printf("Calling clGetPlatformIDs to get number of platforms\n");
    cl_err = clGetPlatformIDs(0, NULL, &num_platforms);
    assert(clError(cl_err, "Getting Platform IDs") == 0);

    printf("Calling clGetPlatformIDs to set platforms\n");
    cl_platform = (cl_platform_id*)malloc(num_platforms*sizeof(cl_platform_id));
    cl_err = clGetPlatformIDs(num_platforms, cl_platform, NULL);
    assert(clError(cl_err, "Getting Platform IDs") == 0);

    printf("Calling clGetDeviceIDs to get number of devices\n");
    cl_err = clGetDeviceIDs(cl_platform[0], CL_DEVICE_TYPE_ALL, 0, NULL, &num_devices);
    assert(clError(cl_err, "Getting Device IDs") == 0);

    printf("Calling clGetDeviceIDs to set devices\n");
    cl_device = (cl_device_id*)malloc(num_devices * sizeof(cl_device_id));
    cl_err = clGetDeviceIDs(cl_platform[0], CL_DEVICE_TYPE_ALL, num_devices, cl_device, NULL);
    assert(clError(cl_err, "Getting Device IDs") == 0);

    printf("Calling clCreateContext\n");
    cl_context ctx = clCreateContext(NULL, 1, &cl_device[DEVICEID], NULL, NULL, &cl_err);
    assert(clError(cl_err, "Creating context") == 0);

    printf("Calling clCreateCommandQueue\n");
    cl_command_queue queue = clCreateCommandQueue(ctx, cl_device[DEVICEID], CL_QUEUE_PROFILING_ENABLE, &cl_err);
    assert(clError(cl_err, "Creating command queue") == 0);


    struct System A;
    int L = 64;
    int N = 2*L;
    int seed = run_throughs; //L+run_throughs+81;
    initSys(&A, L, seed);
    double/**/ *nodes1 = (double/**/*)calloc((L+3)*(N+2), sizeof(nodes1));
    double/**/ *nodes2 = (double/**/*)calloc((L+3)*(N+2), sizeof(nodes2));
    double/**/ *vthresholds = (double/**/*)calloc((L+2)*(N+2), sizeof(vthresholds));
    double/**/ *dthresholds = (double/**/*)calloc((L+2)*(N+2), sizeof(dthresholds));
    int *bonds1 = (int*)calloc((L+3)*(N+2), sizeof(bonds1));
    int *bonds2 = (int*)calloc((L+3)*(N+2), sizeof(bonds2));
    prepareNodes(&A, nodes1, vthresholds, dthresholds);
    memcpy(nodes2, nodes1, (L+3)*(N+2)*sizeof(nodes1));

    int num_wi = L*(L+1);    // Number of work items per work group
    int max_wi = 256;        // Max number of work items per work group allowed
    int max_tot_wi = 71680;  // Max numer of total work items


    num_wi = 64;//max_wi;
    int tot_num_wi = L*(L+1)*2;
    while (tot_num_wi > max_tot_wi)
	tot_num_wi >>= 1;

    printf("\nNum WI: %d\n", num_wi);
    const size_t local_ws = 2*L;//num_wi;//2*L;      // Number of work items per work group
    //const size_t global_ws = (L+1)*num_wi/4; // Total number of work items
    const size_t global_ws = tot_num_wi;       // => global_ws/local_ws = n work groups
    const int num_wg = global_ws/local_ws;     //global_ws/local_ws;
    printf("Work groups: %d \t Items per work group: %lu \t Total number of work items: %lu\n",
	   num_wg, local_ws, global_ws);
    printf("Run %d, size %d\n", seed, L);
    const int nodes_size = (L+3)*(N+2)*sizeof(nodes1);
    const int thresholds_size = (L+2)*(N+2)*sizeof(vthresholds);
    const int bonds_size = (L+3)*(N+2)*sizeof(bonds1);
    const int lsum_size = local_ws*sizeof(cl_double*);

    struct Params params;
    params.size = L;
    params.i_row = 0;
    params.i_col = L/2;
    params.col = 0;
    params.row = 0;
    
    // int rw = 1;

    printf("Creating buffers\n");
    cl_mem d_nodes1 = clCreateBuffer(ctx, CL_MEM_READ_WRITE | CL_MEM_COPY_HOST_PTR,
				     nodes_size, nodes1, &cl_err);
    assert(clError(cl_err, "Creating buffer for nodes1") == 0);

    cl_mem d_nodes2 = clCreateBuffer(ctx, CL_MEM_READ_WRITE | CL_MEM_COPY_HOST_PTR,
				     nodes_size, nodes2, &cl_err);
    assert(clError(cl_err, "Creating buffer for nodes2") == 0);

    cl_mem d_vthresholds = clCreateBuffer(ctx, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR,
					  thresholds_size, vthresholds, &cl_err);
    assert(clError(cl_err, "Creating buffer for vthresholds") == 0);

    cl_mem d_dthresholds = clCreateBuffer(ctx, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR,
					  thresholds_size, dthresholds, &cl_err);
    assert(clError(cl_err, "Creating buffer for dthresholds") == 0);

    cl_mem d_bonds1 = clCreateBuffer(ctx, CL_MEM_WRITE_ONLY | CL_MEM_COPY_HOST_PTR,
				    bonds_size, bonds1, &cl_err);
    assert(clError(cl_err, "Creating buffer for bonds1") == 0);

    cl_mem d_bonds2 = clCreateBuffer(ctx, CL_MEM_WRITE_ONLY | CL_MEM_COPY_HOST_PTR,
				    bonds_size, bonds2, &cl_err);
    assert(clError(cl_err, "Creating buffer for bonds2") == 0);

    cl_mem d_params = clCreateBuffer(ctx, CL_MEM_READ_ONLY,// | CL_MEM_COPY_HOST_PTR, 
    				     sizeof(Params), NULL, &cl_err);
    assert(clError(cl_err, "Creating buffer for params") == 0);
    
    cl_err = clEnqueueWriteBuffer(queue, d_params, CL_TRUE, 0, sizeof(Params),
    				  &params, 0, NULL, NULL);
    assert(clError(cl_err, "Writing buffer to params") == 0);
    // BUFFERS CREATED
    

    char *fsource = load_program_source("optimal_path.cl");

    printf("Creating program\n");
    cl_program program = clCreateProgramWithSource(ctx, 1, (const char**)&fsource, NULL, &cl_err);
    assert(clError(cl_err, "Creating program") == 0);

    printf("Building program\n");
    cl_err = clBuildProgram(program, 1, &cl_device[DEVICEID], "-cl-opt-disable", NULL, NULL);
    if(clError(cl_err, "Building program")){

	// Shows the log
	char* build_log;
	size_t log_size;
	// First call to know the proper size
	clGetProgramBuildInfo(program, cl_device[DEVICEID], CL_PROGRAM_BUILD_LOG, 0, NULL, &log_size);
	build_log = (char*)calloc(log_size+1, sizeof(char));
	// Second call to get the log
	clGetProgramBuildInfo(program, cl_device[DEVICEID], CL_PROGRAM_BUILD_LOG, log_size, build_log, NULL);
	build_log[log_size] = '\0';
	printf("\nBuild log: \n%s\n\n", build_log);
	free(build_log);

	return 1;
    }


    printf("Creating kernel\n");
    
    cl_kernel fracturescape = clCreateKernel(program, "calculateFracturescapeCL", &cl_err);
    assert(clError(cl_err, "Creating kernel") == 0);

    printf("Setting kernel arguments\n");

    cl_err = clSetKernelArg(fracturescape, 0, sizeof(cl_mem), &d_nodes1);
    assert(clError(cl_err, "Setting kernel argument 0") == 0);

    cl_err = clSetKernelArg(fracturescape, 1, sizeof(cl_mem), &d_nodes2);
    assert(clError(cl_err, "Setting kernel argument 1") == 0);

    cl_err = clSetKernelArg(fracturescape, 2, sizeof(cl_mem), &d_bonds1);
    assert(clError(cl_err, "Setting kernel argument 2") == 0);

    cl_err = clSetKernelArg(fracturescape, 3, sizeof(cl_mem), &d_bonds2);
    assert(clError(cl_err, "Setting kernel argument 3") == 0);

    cl_err = clSetKernelArg(fracturescape, 4, sizeof(cl_mem), &d_vthresholds);
    assert(clError(cl_err, "Setting kernel argument 4") == 0);

    cl_err = clSetKernelArg(fracturescape, 5, sizeof(cl_mem), &d_dthresholds);
    assert(clError(cl_err, "Setting kernel argument 5") == 0);

    cl_err = clSetKernelArg(fracturescape, 6, lsum_size*sizeof(cl_double), NULL);
    assert(clError(cl_err, "Setting kernel argument 6") == 0);

    cl_err = clSetKernelArg(fracturescape, 7, sizeof(cl_mem), &d_params);
    assert(clError(cl_err, "Setting kernel argument 7") == 0);
    
    cl_err = clSetKernelArg(fracturescape, 8, sizeof(cl_int), NULL);
    assert(clError(cl_err, "Setting kernel argument 8") == 0);

    // cl_err = clSetKernelArg(fracturescape, 7, sizeof(cl_mem), &d_rw);
    // assert(clError(cl_err, "Setting kernel argument 7") == 0);

    int result;
    size_t size_ret;
    clGetKernelWorkGroupInfo(fracturescape, cl_device[0], CL_KERNEL_WORK_GROUP_SIZE, sizeof(result), &result, &size_ret);
    printf("Maximum WG size: %d\n", result);
    clGetKernelWorkGroupInfo(fracturescape, cl_device[0], CL_KERNEL_PREFERRED_WORK_GROUP_SIZE_MULTIPLE, sizeof(result), &result, &size_ret);
    printf("Preferred WG size: %d\n", result);

    printf("Enqueueing kernel\n");
    double start = walltime();   

    double/**/ *gpu_pathscape = (double/**/*)calloc(L*(L+1), sizeof(gpu_pathscape));
    int **bonds = (int**)calloc(2, sizeof(bonds));
    bonds[0] = (int*)calloc(L*(L+1), sizeof(bonds[0]));
    bonds[1] = (int*)calloc(L*(L+1), sizeof(bonds[1]));

    double/**/ *temp_pathscape = (double/**/*)calloc(L*(L+1), sizeof(temp_pathscape));
    int **temp_bonds = (int**)calloc(2, sizeof(temp_bonds));
    temp_bonds[0] = (int*)calloc(L*(L+1), sizeof(temp_bonds[0]));
    temp_bonds[1] = (int*)calloc(L*(L+1), sizeof(temp_bonds[1]));

    //double/**/ *combined = (double/**/*)calloc(2*L*(L+1), sizeof(combined));
    //double/**/ *combined = (double/**/*)calloc(2*L*(L+1), sizeof(combined));

    for(int i=0; i<L*(L+1); i++)
	gpu_pathscape[i] = DBL_MAX;


    // Each NDRangeKernel calculates the pathscape for one row
    // (i.e. one spanning path between two columns).
    // Loop over all rows (i) and update i_row in params.
    // The pathscape is updated with the partial pathscape
    // before the next row is calculated.
    for(int i=0; i<L+1; i++){
	cl_err = clEnqueueNDRangeKernel(queue, fracturescape, 1, NULL,
					&global_ws, &local_ws, 0, NULL, &event);
	assert(clError(cl_err, "Enqueueing kernel") == 0);

	cl_err = clEnqueueReadBuffer(queue, d_nodes1, CL_TRUE, 0, 
				     nodes_size, nodes1, 0, NULL, NULL);
	assert(clError(cl_err, "Reading buffer nodes1") == 0);

	cl_err = clEnqueueReadBuffer(queue, d_nodes2, CL_TRUE, 0, 
				     nodes_size, nodes2, 0, NULL, NULL);
	assert(clError(cl_err, "Reading buffer nodes2") == 0);       

	writeToGPUFracturescape(L, gpu_pathscape, bonds, nodes1, nodes2, bonds1, bonds2, temp_pathscape, temp_bonds);

	params.i_row += 1;
	cl_err = clEnqueueWriteBuffer(queue, d_params, CL_TRUE, 0, sizeof(Params),
				      &params, 0, NULL, NULL);
	assert(clError(cl_err, "Writing buffer to params") == 0);

	cl_err = clSetKernelArg(fracturescape, 7, sizeof(cl_mem), &d_params);
	assert(clError(cl_err, "Setting kernel argument 7") == 0);


    }

    clFinish(queue);



    clGetEventProfilingInfo(event, CL_PROFILING_COMMAND_START,
			    sizeof(cl_ulong), &start_time, NULL);
    clGetEventProfilingInfo(event, CL_PROFILING_COMMAND_END,
			    sizeof(cl_ulong), &end_time, NULL);
    cl_ulong kernelExecTimeNs = end_time-start_time;
    
    printf("Walltime, gpu: %.6f \t Eventtime, gpu: %.6f\n", walltime()-start, kernelExecTimeNs/1e9);


    cl_err = clEnqueueReadBuffer(queue, d_nodes1, CL_TRUE, 0, 
    				 nodes_size, nodes1, 0, NULL, NULL);
    assert(clError(cl_err, "Reading buffer nodes1") == 0);

    cl_err = clEnqueueReadBuffer(queue, d_nodes2, CL_TRUE, 0, 
    				 nodes_size, nodes2, 0, NULL, NULL);
    assert(clError(cl_err, "Reading buffer nodes2") == 0);
    
    free(cl_platform);
    free(cl_device);


    // printf("\n\nPrinting gpu pathscape...\n\n");
    // for(int i=0; i<L+1; i++){
    // 	for(int j=0; j<L; j++)
    //	    printf("%.3f  ", gpu_pathscape[i*L+j]);
    //	printf("\n");
    // }

    // printf("\n\nPrinting nodes1...\n\n");
    // for(int i=0; i<L+3; i++){
    // 	for(int j=0; j<N+2; j++)
    // 	    printf("%.0f  ", nodes1[i*(N+2)+j]);
    // 	printf("\n");
    // }
    
    // printf("\n\nPrinting nodes2...\n\n");
    // for(int i=1; i<L+2; i++){
    // 	for(int j=1; j<N+1; j++)
    // 	    printf("%.1f  ", nodes2[i*(N+2)+j]);
    // 	printf("\n");
    // }


    start = walltime();
    findFracturescape(&A);
    printf("\nWalltime cpu: %.6f\n", walltime()-start);

    // printf("\n\nPrinting pathscape...\n\n");
    // for(int i=0; i<L+1; i++){
    // 	for(int j=0; j<L; j++)
    // 	    printf("%.4f  ", A.pathscape[i*L+j]);
    // 	printf("\n");
    // }

    int errors = 0;
    for(int i=0; i<L+1; i++)
    	for(int j=0; j<L; j++)
	    errors += fabs(A.pathscape[i*L+j]-gpu_pathscape[i*L+j]) > 1e-6;
    printf("Disrepancies between cpu and gpu results: %d\n\n", errors);


    printf("Cleaning up...\n");

    clReleaseKernel(fracturescape);
    clReleaseCommandQueue(queue);
    clReleaseContext(ctx);
    clReleaseMemObject(d_nodes1);
    clReleaseMemObject(d_nodes2);
    clReleaseMemObject(d_vthresholds);
    clReleaseMemObject(d_dthresholds);
    clReleaseMemObject(d_bonds1);
    clReleaseMemObject(d_bonds2);
    // clReleaseMemObject(d_rw);
    clReleaseMemObject(d_params);

    free(gpu_pathscape);
    free(temp_pathscape);
    free(nodes1);
    free(nodes2);
    free(vthresholds);
    free(dthresholds);
    free(bonds1);
    free(bonds2);
    free(bonds[0]);
    free(bonds[1]);
    free(bonds);
    free(temp_bonds[0]);
    free(temp_bonds[1]);
    free(temp_bonds);
    
	
    freeSys(&A);

    }

    return 0;
}

